/* public domain is this file */

#include <string.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "os_dep.h"

#define TEXX 32
#define TEXY 32

static void init_boule(unsigned char boule[TEXX][TEXY],
		       unsigned char intensite[256][256],
		       unsigned char sature[256][256])
{
  register int i, j;

  //#define MAX (sqrt(128*128*0.8)-12)
  #define MAX (sqrt((TEXX/2)*(TEXY/2)*0.8)-70/32)
  for (i=0; i<TEXX; i++)
    for (j=0; j<TEXY; j++) {
      double z=(MAX-sqrt((((double)TEXX/2.)-(double)i)*(((double)TEXX/2.)-(double)i)+
			 (((double)TEXY/2.)-(double)j)*0.8*0.8*(((double)TEXY/2.)-(double)j)))*255/MAX;
      if (z<0)
	boule[i][j]=0;
      else
	boule[i][j]=z;
    }

  for (i=0; i<256; i++)
    for (j=0; j<256; j++)
      intensite[i][j]=(double)i*(double)j/255.;

  for (i=0; i<256; i++)
    for (j=0; j<256; j++)
      if (i>j)
	sature[i][j]=i+j>255?255:i+j;//+j/4>255?255:i+j/4+5>255?255:i+j/4+5;
      else
	sature[i][j]=j+i>255?255:i+j;//+i/4>255?255:j+i/4+5>255?255:j+i/4+5;
}

static void put_boule(int intens, int x, int y,
		      unsigned char boule[TEXX][TEXY],
		      unsigned char intensite[256][256],
		      unsigned char sature[256][256])
{
  register int i, j;
  register unsigned char *p=&boule[0][0], *t=ecran1, *inten=&intensite[intens][0];
  register int incp=0, inct=320-TEXX, nblignes=TEXY, nbcols=TEXX;

  if (intens<20)
    return;

  x-=TEXX/2;
  y-=TEXY/2;
  if (x<-(TEXX-1) || x>=320 || y<-(TEXY-1) || y>=200)
    return;

  t+=x;
  t+=y*320;

  if (*(t+320*16+16)>80)
    return;

  nblignes=200-y<TEXY?200-y:TEXY;
  if (nblignes>200)
    nblignes=200;

  if (x<0) {
    t-=x;
    p-=x;
    nbcols+=x;
    incp=-x;
    inct-=x;
  }
  if (x>=320-TEXX) {
    x=x+TEXX-320;
    nbcols-=x;
    inct+=x;
    incp=+x;
  }
  if (y<0) {
    t-=y*320;
    p-=y*TEXX;
    nblignes=TEXY+y;
    if (nblignes>200)
      nblignes=200;
  }

  for (i=0; i<nblignes; i++, t+=inct, p+=incp)
    for (j=0; j<nbcols; j++, p++, t++)
      *t=sature[*t][inten[*p]];
}

static void clear_screen(void)
{
  memset(ecran1, 0, 64000);
}

static void init_palette(void)
{
  unsigned char pal[256][3];
  int i;

  for (i=0; i<256; i++)
    pal[i][0]=pal[i][1]=pal[i][2]=i;
  put_palette(pal);
}

#define Xpoint 100
#define Ypoint 100
#define RSUN	57
#define RMOON	47
int nbpoints;

static void init_points(int points[Xpoint*Ypoint][2])
{
  int i, j;
  nbpoints=0;
  for (i=0; i<Xpoint; i++) {
    double x=(double)i/(double)(Xpoint-1)*319.;
    double dx=x-160;
    dx*=dx*0.8;
    for (j=0; j<Ypoint; j++) {
      double y=(double)j/(double)(Ypoint-1)*199.;
      double dy=y-100;
      dy*=dy;
      if (dx+dy<(double)RSUN*(double)RSUN*0.9) {
	/* let's put the point */
	points[nbpoints][0]=x;
	points[nbpoints++][1]=y;
      }
    }
  }
  fprintf(stdout, "We've got %d points in the sun\n", nbpoints);
}

static void show_points(double x, double y,
			unsigned char boule[TEXX][TEXY],
			unsigned char intensite[256][256],
			unsigned char sature[256][256],
			int points[Xpoint*Ypoint][2])
{
  int i;
  int dx, dy;
  int firstpoint;

  firstpoint=(double)rand()/(RAND_MAX+1.)*(double)nbpoints;
  #define VIBRE 10.
  dx=(double)rand()/(RAND_MAX+1.)*VIBRE-VIBRE/2;
  dy=(double)rand()/(RAND_MAX+1.)*VIBRE-VIBRE/2;
  for (i=firstpoint; i<nbpoints; i++) {
    #define X ((double)points[i][0]-x)
    #define Y ((double)points[i][1]-y)
    double d=X*X*0.8+Y*Y-(double)RMOON*(double)RMOON;
    if (d<0)
      d=0;
    d=sqrt(d);
    if (d>255)
      d=255;
    put_boule(d, points[i][0]+dx, points[i][1]+dy,
	      boule, intensite, sature);
  }

  for (i=0; i<firstpoint; i++) {
    double d=X*X*0.8+Y*Y-(double)RMOON*(double)RMOON;
    if (d<0)
      d=0;
    d=sqrt(d);
    if (d>255)
      d=255;
    put_boule(d, points[i][0]+dx, points[i][1]+dy,
	      boule, intensite, sature);
  }
}

void main_eclipse(void)
{
  unsigned char boule[TEXX][TEXY];
  unsigned char intensite[256][256];
  unsigned char sature[256][256];
  int points[Xpoint*Ypoint][2];
  int nb_screen=0;
  double time=0.;

  init_palette();
  init_boule(boule, intensite, sature);
  init_points(points);
  reset_timer(10.*1000.);
  while (time<=1.) {
    clear_screen();
    /*
      0	   t    1
      XMIN x    XMAX
      t=(x-XMIN)/(XMAX-XMIN) => x=t*(XMAX-XMIN)+XMIN

      0		1
      YMAX	YMIN
     */
    #define XMIN -50
    #define XMAX (320+50)
    #define YMIN -50
    #define YMAX 250
    show_points(XMIN+time*(XMAX-XMIN),YMAX+time*(YMIN-YMAX),
		boule, intensite, sature, points);
    wait_retrace();
    put_screen();
    nb_screen++;
    time=get_time();
  }
  fprintf(stderr, "eclipse part : nb screen=%d for 10s => %f fps\n", nb_screen, (double)nb_screen/10.);
}
