bits 32
section .text

extern ecran1
extern xpt
extern ypt
extern _uuu
extern _uu
extern _u

global interpolate

;Thanks to Cyg/Blabla for his good explaination of a good way to do bilinear interpolation.
;You can find it to http://blabla.planet-d.net, just read it and understand it and use it !
;
;I rewrite it here for it's very simple and useful.
;
; You have calculated something to 4 points (v1, v2, v3, v4)
;and you want to put intermediate values v(x, y) inside the square
;(you do that if the calculation of the values take a lot of cpu)
; You choose to say that it will be linearly interpolated (because
;it's a simple way to interpolated and in most of case, even if this
;way of interpolate is not exact, you cannot see the difference).
;
; That gives the picture :
;
;  -------------------------------> X axis
; |
; |
; |   v1 ----------- v2
; |   |               |
; |   |               |
; |   |               |
; |   |   *           |
; |   |    v(x,y)     |
; |   |               |
; |   |               |
; |   |               |
; |   v3 ----------- v4
; |
; |
; |
; V Y axis
;
;
; So, the v point has the value :
;
; v = (1-x)(1-y) * v1 + x(1-y) * v2 + (1-x)y * v3 + xy * v4
;
; This is the definition of bilinear interpolatation.
; The value is correct for 0<=x<=1 and 0<=y<=1
;
; Now, as you see, there are 8 multiplications. This consumes a lot of cpu,
;in fixed-point math or in floation-point. In demo (realtime applications)
;we want a quick and fast way to calculate.
; So we cannot use this formula as it is (well, if our function requires
;more than 3 muls to calculate, we still win some cpu, but we can do much
;better !)
; So how do we do ?
; Well, we have the answer. This is the derivation of the function.
;
; What's this all ?
; Well, derivation is basic mathematic that you learn to school.
;It is used in physic where you want to calculate speed from positions
;plot of something moving.
; Well, if you don't know what it is, go learn. I can't do it for you.
;
; Derivation is here used to calculate values at points we don't want to
;calculate the function.
; The fundamental equation with derivation is :
; d(f(x))/dx=lim(h->0) (f(x+h)-f(x))/h.
; Let's suppose a small h (we can do it there, because our function is a bilinear one
;with x and y in [0..1]. But of course it's not very perfect, but that's enough for
;what we do).
; So we can write (forgetting the limit) :
; D(f(x))/Dx=(f(x+h)-f(x))/h
;or :
; f(x+h)=f(x)+D(f(x))/Dx * h
;So, to calculate f(x+h) you only need to know f(x) and D(f(x))/Dx.
;Step by step, you calculate all the f(x) you want, but with only one f(x) !
;If D(f(x))/Dx is easy to calculate, you win. Yes you have a multiplication,
;but you will see it's done only one time in the sqare, and if you have
;something multiple of 2, you do a sar or sal, which is ok (so, you'll use
;fixed point math).
;
; The equation given is for a total derivation with one variable.
; For several variables (as we have here) we do partial derivation,
;with the equation (guess 3 variables and derivating from the first one):
; d(f(x,y,z))/dx=lim(h->0) (f(x+h,y,z)-f(x,y,z))/h.
;Then we've got the same than for a total derivation.
;
; In our case, the function is f(x,y)=(1-x)(1-y) * v1 + x(1-y) * v2 + (1-x)y * v3 + xy * v4
;Mathematical theory say to us that the partial derivations from x and y are exactly :
;
; df/dx = -(1-y) * v1 + (1-y) * v2 - y * v3 + y * v4
;
; df/dy = -(1-x) * v1 - x * v2 + (1-x) * v3 + x * v4
;
; So now, guess we are in v1 and we want to calculate values up to v2.
;Let's say we have 8 values from v1 to v2 (i mean that in your demo you calculate
;the point v1 to the screen then the point v2 where the x of v2 is x of v1 + 8, and
;you do the same in the y direction for v1 and v3, that's very common for hard function
;that require lot's of cpu, as I already said).
; Because you walk through x, you have a constant df/dx (it only depends of y as you can see,
;which is great for us, that means we have a constant ! no calculation needed for df/dx, good
;he ?), so at each step you add df/dx * h (with h = 1/8 because we take 8 values).
;
; Then you do the same between v1 and v3. So you'll have y from v1 + 1/8 that is the new
;starting point where you'll have to interpolate to go to y from v2 + 1/8 (you get this
;new point by adding df/dy * h to v1).
; So y has changed, so you think you have to calculate another df/dx for this line.
; Yes, you do, for sure. But in df/dx has 4 muls, which is big. So what we do is to
;take df/dx for another function that we derivate a new time as we derivated f. This
;time, we run through it with y value. So we simply derivate it with y and there we
;go again ! Only one add to have the new value (I am not sure if this is all correct,
;but in practice it's fine, so...)
; So we have :
;
; df/dxdy = v1 - v2 - v3 + v4
;
; But there, we'll have to multiply by (h*h) because it's a double derivation
;(well, don't ask me more about it, that's how Cyg does, so it must be right !)
;
; You get it ? It's very simple in fact. I understood Cyg's article, so if I did, you can do it !
;
; So, you'll have a routine that looks this way (for an 8 points stuff) :
; (I use a 4 points one so I do h=1/4. Well, in my code, I do fixed math
;with 8 bits "apres la virgule", so only shl are needed.)
; h= 1/8
; sv = v1
; dfdx = (-v1 + v2) * h (we start in v1, so x=y=0)
; dfdy = (-v1 + v3) * h
; dfdxdy = (v1 - v2 - v3 + v4) * (h * h)
; for y = 0 to 7 do
;   v=sv
;   for x = 0 to 7 do
;     put v at its place
;     v += dfdx
;   done
;   sv += dfdy
;   dfdx += dfdxdy
; done
;
; As you see, you only have 3 muls, and because it's an 8x8 square, that's not mul but sar
;which is must better ! You win a lot with this method !
;
; Of course, in asm, you won't have the for, so your routine will be quick and use less
;registers which is good for our good old 8-registers PC, nope ? Have fun with all this.
;
; The main use is to calculate 2d effects 1 point on 8 in x and in y then interpolate the
;intermediate values (for this production, I calculate points in a texture so my v is in
;fact a vector).
;
; Very special thanks to Cyg for letting me understand it !
;

align 16

%define dfdx1		dword [ebp-12]
%define dfdx2		dword [ebp-16]
%define dfdy1		dword [ebp-20]
%define dfdy2		dword [ebp-24]
%define dfdxdy1		dword [ebp-28]
%define dfdxdy2		dword [ebp-32]
%define sv1		dword [ebp-36]
%define sv2		dword [ebp-40]

%macro do_interpolate 0
	mov	eax, [esi*4+xpt]
	mov	ebx, [esi*4+ypt]
	mov	ecx, eax
	mov	edx, ebx
	shl	eax, 8
	shl	ebx, 8
	mov	sv1, eax
	mov	sv2, ebx
	mov	eax, ecx
	mov	ebx, edx
	neg	eax
	neg	ebx
	add	eax, [esi*4+xpt+4*4]
	add	ebx, [esi*4+ypt+4*4]
	shl	eax, 6
	shl	ebx, 6
	mov	dfdx1, eax
	mov	dfdx2, ebx
	mov	eax, ecx
	mov	ebx, edx
	neg	eax
	neg	ebx
	add	eax, [esi*4+xpt+320*4*4]
	add	ebx, [esi*4+ypt+320*4*4]
	sub	ecx, [esi*4+xpt+320*4*4]
	sub	edx, [esi*4+ypt+320*4*4]
	shl	eax, 6
	shl	ebx, 6
	mov	dfdy1, eax
	mov	dfdy2, ebx
	add	ecx, [esi*4+xpt+(320*4+4)*4]
	add	edx, [esi*4+ypt+(320*4+4)*4]
	sub	ecx, [esi*4+xpt+4*4]
	sub	edx, [esi*4+ypt+4*4]
	shl	ecx, 4
	shl	edx, 4
	mov	dfdxdy1, ecx
	mov	dfdxdy2, edx

	mov	eax, sv1
	mov	ebx, sv2
	mov	edx, [ebp+8]
	xor	ecx, ecx
	;pt (0, 0) (at its place)
	;pt (1, 0) (2, 0) (3, 0)
	add	eax, dfdx1
	add	ebx, dfdx2
	mov	cl, bh
	mov	ch, ah
	mov	cl, [ecx+edx]
	mov	[esi+1+edi], cl
	add	eax, dfdx1
	add	ebx, dfdx2
	mov	cl, bh
	mov	ch, ah
	mov	cl, [ecx+edx]
	mov	[esi+2+edi], cl
	add	eax, dfdx1
	add	ebx, dfdx2
	mov	cl, bh
	mov	ch, ah
	mov	cl, [ecx+edx]
	mov	[esi+3+edi], cl
	;pt (0, 1), (1, 1), (2, 1), (3, 1)
	mov	eax, dfdx1
	mov	ebx, dfdx2
	add	eax, dfdxdy1
	add	ebx, dfdxdy2
	mov	dfdx1, eax
	mov	dfdx2, ebx
	mov	eax, sv1
	mov	ebx, sv2
	add	eax, dfdy1
	add	ebx, dfdy2
	mov	sv1, eax
	mov	sv2, ebx
	mov	cl, bh
	mov	ch, ah
	mov	cl, [ecx+edx]
	mov	[esi+320+0+edi], cl
	add	eax, dfdx1
	add	ebx, dfdx2
	mov	cl, bh
	mov	ch, ah
	mov	cl, [ecx+edx]
	mov	[esi+320+1+edi], cl
	add	eax, dfdx1
	add	ebx, dfdx2
	mov	cl, bh
	mov	ch, ah
	mov	cl, [ecx+edx]
	mov	[esi+320+2+edi], cl
	add	eax, dfdx1
	add	ebx, dfdx2
	mov	cl, bh
	mov	ch, ah
	mov	cl, [ecx+edx]
	mov	[esi+320+3+edi], cl
	;pt (0, 2), (1, 2), (2, 2), (3, 2)
	mov	eax, dfdx1
	mov	ebx, dfdx2
	add	eax, dfdxdy1
	add	ebx, dfdxdy2
	mov	dfdx1, eax
	mov	dfdx2, ebx
	mov	eax, sv1
	mov	ebx, sv2
	add	eax, dfdy1
	add	ebx, dfdy2
	mov	sv1, eax
	mov	sv2, ebx
	mov	cl, bh
	mov	ch, ah
	mov	cl, [ecx+edx]
	mov	[esi+320*2+0+edi], cl
	add	eax, dfdx1
	add	ebx, dfdx2
	mov	cl, bh
	mov	ch, ah
	mov	cl, [ecx+edx]
	mov	[esi+320*2+1+edi], cl
	add	eax, dfdx1
	add	ebx, dfdx2
	mov	cl, bh
	mov	ch, ah
	mov	cl, [ecx+edx]
	mov	[esi+320*2+2+edi], cl
	add	eax, dfdx1
	add	ebx, dfdx2
	mov	cl, bh
	mov	ch, ah
	mov	cl, [ecx+edx]
	mov	[esi+320*2+3+edi], cl
	;pt (0, 3), (1, 3), (2, 3), (3, 3)
	mov	eax, dfdx1
	mov	ebx, dfdx2
	add	eax, dfdxdy1
	add	ebx, dfdxdy2
	mov	dfdx1, eax
	mov	dfdx2, ebx
	mov	eax, sv1
	mov	ebx, sv2
	add	eax, dfdy1
	add	ebx, dfdy2
	mov	sv1, eax
	mov	sv2, ebx
	mov	cl, bh
	mov	ch, ah
	mov	cl, [ecx+edx]
	mov	[esi+320*3+0+edi], cl
	add	eax, dfdx1
	add	ebx, dfdx2
	mov	cl, bh
	mov	ch, ah
	mov	cl, [ecx+edx]
	mov	[esi+320*3+1+edi], cl
	add	eax, dfdx1
	add	ebx, dfdx2
	mov	cl, bh
	mov	ch, ah
	mov	cl, [ecx+edx]
	mov	[esi+320*3+2+edi], cl
	add	eax, dfdx1
	add	ebx, dfdx2
	mov	cl, bh
	mov	ch, ah
	mov	cl, [ecx+edx]
	mov	[esi+320*3+3+edi], cl
%endmacro

interpolate:
%define x dword [ebp-4]
%define y dword [ebp-8]
	push	ebp
	mov	ebp, esp
	sub	esp, 40
	push	edi
	push	ebx
	push	esi

	;mov	ebp, [ebp+8]
	mov	edi, [ecran1]
	xor	esi, esi
	mov	y, 200/4-1
.01:	mov	x, 320/4-1
.02:	do_interpolate
	add	esi, 4
	dec	x
	jnz	near .02
	add	esi, 4+320*3
	dec	y
	jnz	near .01

	pop	esi
	pop	ebx
	pop	edi
	add	esp, 40
	pop	ebp
	ret

%if 0
	mov	esi, 1
	mov	ecx, (200-3)/2
.1:	mov	edx, (319-3)/2-1
.2:	mov	eax, [esi*4+ypt-1*4]
	mov	ebx, [esi*4+xpt-1*4]
	add	eax, [esi*4+ypt+1*4]
	add	ebx, [esi*4+xpt+1*4]
	sar	eax, 1
	sar	ebx, 1
	mov	[esi*4+ypt], eax
	mov	[esi*4+xpt], ebx
	and	eax, 255
	and	ebx, 255
	mov	ah, bl
	mov	al, [eax+ebp]
	mov	[edi+esi], al
	add	esi, 2
	dec	edx
	jns	.2
	add	esi, 4+320
	dec	ecx
	jns	.1

	mov	esi, 320
	mov	ecx, (200-3)/2-1
.7:	mov	edx, (319-3)
.8:	mov	eax, [esi*4+ypt-(320)*4]
	mov	ebx, [esi*4+xpt-(320)*4]
	add	eax, [esi*4+ypt+(320)*4]
	add	ebx, [esi*4+xpt+(320)*4]
	sar	eax, 1
	sar	ebx, 1
	;mov	[esi*4+ypt], eax
	;mov	[esi*4+xpt], ebx
	and	eax, 255
	and	ebx, 255
	mov	ah, bl
	mov	al, [eax+ebp]
	mov	[edi+esi], al
	inc	esi
	dec	edx
	jns	.8
	add	esi, 3+320
	dec	ecx
	jns	.7
%endif
%if 0
%if 0
	mov	esi, 2
	mov	ecx, (200-3)/4
.1:	mov	edx, (319-3)/4-1
.2:	mov	eax, [esi*4+ypt-2*4]
	mov	ebx, [esi*4+xpt-2*4]
	add	eax, [esi*4+ypt+2*4]
	add	ebx, [esi*4+xpt+2*4]
	sar	eax, 1
	sar	ebx, 1
	mov	[esi*4+ypt], eax
	mov	[esi*4+xpt], ebx
	and	eax, 255
	and	ebx, 255
	mov	ah, bl
	mov	al, [eax+ebp]
	mov	[edi+esi], al
	add	esi, 4
	dec	edx
	jns	.2
	add	esi, 4+320*3
	dec	ecx
	jns	.1
%endif
	mov	esi, 1
	mov	ecx, (200-3)/4
.3:	mov	edx, (319-3)/2-1
.4:	mov	eax, [esi*4+ypt-1*4]
	mov	ebx, [esi*4+xpt-1*4]
	add	eax, [esi*4+ypt+1*4]
	add	ebx, [esi*4+xpt+1*4]
	sar	eax, 1
	sar	ebx, 1
	mov	[esi*4+ypt], eax
	mov	[esi*4+xpt], ebx
	and	eax, 255
	and	ebx, 255
	mov	ah, bl
	mov	al, [eax+ebp]
	mov	[edi+esi], al
	add	esi, 2
	dec	edx
	jns	.4
	add	esi, 4+320*3
	dec	ecx
	jns	.3

	mov	esi, 320*2
	mov	ecx, (200-3)/4-1
.5:	mov	edx, (319-3)
.6:	mov	eax, [esi*4+ypt-(320*2)*4]
	mov	ebx, [esi*4+xpt-(320*2)*4]
	add	eax, [esi*4+ypt+(320*2)*4]
	add	ebx, [esi*4+xpt+(320*2)*4]
	sar	eax, 1
	sar	ebx, 1
	mov	[esi*4+ypt], eax
	mov	[esi*4+xpt], ebx
	and	eax, 255
	and	ebx, 255
	mov	ah, bl
	mov	al, [eax+ebp]
	mov	[edi+esi], al
	inc	esi
	dec	edx
	jns	.6
	add	esi, 3+320*3
	dec	ecx
	jns	.5

	mov	esi, 320
	mov	ecx, (200-3)/2-1
.7:	mov	edx, (319-3)
.8:	mov	eax, [esi*4+ypt-(320)*4]
	mov	ebx, [esi*4+xpt-(320)*4]
	add	eax, [esi*4+ypt+(320)*4]
	add	ebx, [esi*4+xpt+(320)*4]
	sar	eax, 1
	sar	ebx, 1
	;mov	[esi*4+ypt], eax
	;mov	[esi*4+xpt], ebx
	and	eax, 255
	and	ebx, 255
	mov	ah, bl
	mov	al, [eax+ebp]
	mov	[edi+esi], al
	inc	esi
	dec	edx
	jns	.8
	add	esi, 3+320
	dec	ecx
	jns	.7
%endif
%if 0
	pop	esi
	pop	ebx
	pop	edi
	pop	ebp
	ret
%endif

%if 1
global bezier
align 16
bezier:
%define pts_tab	[ebp+8]
%define l	[ebp+12]
%define texture	[ebp+16]
%define a1	[ebp-4]
%define b1	[ebp-8]
%define c1	[ebp-12]
%define d1	[ebp-16]
%define a2	[ebp-20]
%define b2	[ebp-24]
%define c2	[ebp-28]
%define d2	[ebp-32]
%define dum	[ebp-36]
%define t(i,j)	[esi+i*8+j*4]
	push	ebp
	mov	ebp, esp
	sub	esp, 8*4+4
	push	esi
	push	ebx
	push	edi

	mov	esi, pts_tab

	mov	eax, t(3,0)
	mov	ebx, t(1,0)
	mov	ecx, t(2,0)
	lea	ebx, [ebx+ebx*2]
	add	eax, ebx
	lea	ecx, [ecx+ecx*2]
	sub	eax, ecx
	sub	eax, t(0,0)
	mov	a1, eax

	mov	eax, t(0,0)
	mov	ebx, t(1,0)
	lea	eax, [eax+eax*2]
	lea	ebx, [ebx+ebx*2]
	shl	ebx,1
	mov	ecx, t(2,0)
	sub	eax, ebx
	lea	ecx, [ecx+ecx*2]
	add	eax, ecx
	mov	b1, eax

	mov	eax, t(1,0)
	mov	ebx, t(0,0)
	lea	eax, [eax+2*eax]
	lea	ebx, [ebx+2*ebx]
	sub	eax, ebx
	mov	c1, eax

	mov	eax, t(0,0)
	mov	d1, eax

	mov	eax, t(3,1)
	mov	ebx, t(1,1)
	mov	ecx, t(2,1)
	lea	ebx, [ebx+ebx*2]
	add	eax, ebx
	lea	ecx, [ecx+ecx*2]
	sub	eax, ecx
	sub	eax, t(0,1)
	mov	a2, eax

	mov	eax, t(0,1)
	mov	ebx, t(1,1)
	lea	eax, [eax+eax*2]
	lea	ebx, [ebx+ebx*2]
	shl	ebx,1
	mov	ecx, t(2,1)
	sub	eax, ebx
	lea	ecx, [ecx+ecx*2]
	add	eax, ecx
	mov	b2, eax

	mov	eax, t(1,1)
	mov	ebx, t(0,1)
	lea	eax, [eax+2*eax]
	lea	ebx, [ebx+2*ebx]
	sub	eax, ebx
	mov	c2, eax

	mov	eax, t(0,1)
	mov	d2, eax

	mov	edi, l
	lea	edi, [edi+edi*4]
	shl	edi, 6
	mov	esi, [ecran1]
	add	esi, edi
	shl	edi, 2

	xor	ebx, ebx
.1:	mov	eax, a1
	mov	ecx, b1

	mov	edx, c1

	imul	eax, [_uuu+ebx*4]

	imul	ecx, [_uu+ebx*4]

	imul	edx, [_u+ebx*4]

	add	eax, ecx

	add	eax, edx

	sar	eax, 10

	add	eax, d1

	mov	[xpt+edi+ebx*4], eax

	mov	dum, eax

	mov	eax, a2

	mov	ecx, b2
	mov	edx, c2

	imul	eax, [_uuu+ebx*4]

	imul	ecx, [_uu+ebx*4]

	imul	edx, [_u+ebx*4]

	add	eax, ecx

	add	eax, edx

	sar	eax, 10

	add	eax, d2

	mov	[ypt+edi+ebx*4], eax
	and	eax, 255

	mov	ah, dum

	add	eax, texture

	mov	al, [eax]

	mov	[esi+ebx], al

	add	ebx, 4

	cmp	ebx, 319
	jbe	near .1

	pop	edi
	pop	ebx
	pop	esi
	add	esp, 8*4+4
	pop	ebp
	ret
%endif
