/* this file is public domain */

#include "parts.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "os_dep.h"
#include "palette.h"
#include "sub_parts.h"
#include "image.h"
#include "cube.h"
#include "point.h"
#include "fixed_math.h"
#include "donut.h"
#include "fillpoly.h"
#include "sound.h"

/******************************************/
/* the title part routines */
void sub_title(char *n, char b, double t)
{
  double time=0;
  image im;

  if (load_image(&im, n)==-1) {
    fprintf(stderr, "mmm, I can't load %s, sorry no sub-title part\n", n);
    return;
  }
#ifdef DEBUG
  {
    int i;
    for (i=0; i<16; i++)
      fprintf(stderr, "%2.2d = %3.3d %3.3d %3.3d\n", i, (unsigned char)im.palette[i*3+0],
	      (unsigned char)im.palette[i*3+1], (unsigned char)im.palette[i*3+2]);
  }
#endif
  reset_palette();
  sub_palette(im.palette, 0, im.palsize);
  save_palette();
  fade_palette(255, 255, 255, 0., 0, 255);
  set_palette();
  clear_screen(b);
  put_image(&im, (320-im.width)/2, (200-im.height)/2);
  put_screen();
  reset_timer(t*1000.0);
  while(time<.2) {
    fade_palette(255, 255, 255, time/.2, 0, 255);
    set_palette();
    time=get_time();
  }
  restore_palette();
  while(time<.8)
    time=get_time();
  while(time<1.) {
    fade_palette(255, 255, 255, 1.-(time-0.8)/.2, 0, 255);
    set_palette();
    time=get_time();
  }
  free_image(&im);
}

void title(void)
{
  sub_title("graphs/title1.png", 8, 3.5/*4.*/);
  sub_title("graphs/title2.png", 7, 3.5/*6.*/);
  pause_me(2.3/*3.*/*1000.);
  sub_title("graphs/title3.png", 8, 5.5);
}

/******************************************/
/* sinus scroll part */
int sinpos[320];

void put_colonne(int x, int y, unsigned char *data)
{
  unsigned char *p=&ecran1[y*320+x];
  int i;
  for (i=0; i<32; i++, p+=320, data--)
    if (*data!=0)
      *p=*data+16*3-1;
}

void dosin(double time, image *im)
{
  int i;
  int start=time*(1280.-320.);
  for (i=0; i<320; i++) {
    int v=100-16+cos((double)i/(40.)-time*50.)*40;
    sinpos[i]=v;
    /* bad this cos, where is the costab ?? */
    put_colonne(i, v, &im->data[(i+start)*32+31]);
  }
}

void sin_clear_screen(unsigned char *sc)
{
  int i;
  for (i=0; i<320; i++) {
    int v=i+sinpos[i]*320;
    unsigned char *e=ecran1+v;
    unsigned char *f=sc+v;
    int j;
    for (j=0; j<32; j++, e+=320, f+=320)
      *e=*f;
  }
}

void sub_sin(void)
{
  #define SINT 10.
  double time=0;
  char sc[64000];
  image im;
  if (load_image(&im, "graphs/sinscroll.png")==-1) {
    fprintf(stderr, "sinus scroll part error when loading \"%s\"...\n", "graphs/sinscroll.png");
    return;
  }
#ifdef DEBUG
  {
    int i;
    for (i=0; i<16; i++)
      fprintf(stderr, "%2.2d = %3.3d %3.3d %3.3d\n", i, (unsigned char)im.palette[i*3+0],
	      (unsigned char)im.palette[i*3+1], (unsigned char)im.palette[i*3+2]);
  }
#endif

  memcpy(sc, ecran1, 64000);
  sub_palette((unsigned char *)&im.palette[3], 16*3, 16);
  set_palette();
  save_palette();

  reset_timer(SINT*1000.);
  while(time<(SINT-2.)/SINT) {
    dosin(time, &im);
    put_screen();
    sin_clear_screen(sc);
    //memcpy(ecran1, sc, 64000);
    time=get_time();
  }
  while(time<1.) {
    dosin(time, &im);
    put_screen();
    sin_clear_screen(sc);
    //memcpy(ecran1, sc, 64000);
    fade_palette(255, 255, 255, 1.-(time-(SINT-2)/SINT)*(SINT/2.), 0, 255);
    set_palette();
    time=get_time();
  }

  free_image(&im);
}

void sinus_scroll(void)
{
  if (show3("graphs/scroll1.png", (320-239)/2, 10,
	    "graphs/scroll2.png", (320-198)/2, 10+35+10,
	    "graphs/scroll3.png", (320-115)/2, 10+35+10+72+10, 7)==-1)
    return;
  pause_me(2.*1000.);
  sub_sin();
}

/******************************************/
/* color bars */
int barpos[4];

void bar(int nb, double angle, int basecol)
{
  char *p=ecran1;
  int d=(int)(cos(angle)*50+100-4)*320+50;
  p+=d;
  barpos[nb]=d;
  memset(p, basecol+3, 320-100);
  p+=320;
  memset(p, basecol+2, 320-100);
  p+=320;
  memset(p, basecol+1, 320-100);
  p+=320;
  memset(p, basecol, 320-100);
  p+=320;
  memset(p, basecol, 320-100);
  p+=320;
  memset(p, basecol+1, 320-100);
  p+=320;
  memset(p, basecol+2, 320-100);
  p+=320;
  memset(p, basecol+3, 320-100);
}

void dobars(double time)
{
  /* bad, should use a cos-table, huh ? */
  #define BARSV 8.
  double angle=time*70.;
  bar(0, angle-3*M_PI/BARSV, 16*3+12);
  bar(1, angle-2*M_PI/BARSV, 16*3+8);
  bar(2, angle-M_PI/BARSV, 16*3+4);
  bar(3, angle, 16*3);
}

void bar_clear_screen(unsigned char *sc)
{
  int i;
  for (i=0; i<4; i++) {
    unsigned char *d=ecran1+barpos[i];
    unsigned char *e=sc+barpos[i];
    memcpy(d, e, 320-100);
    d+=320;
    e+=320;
    memcpy(d, e, 320-100);
    d+=320;
    e+=320;
    memcpy(d, e, 320-100);
    d+=320;
    e+=320;
    memcpy(d, e, 320-100);
    d+=320;
    e+=320;
    memcpy(d, e, 320-100);
    d+=320;
    e+=320;
    memcpy(d, e, 320-100);
    d+=320;
    e+=320;
    memcpy(d, e, 320-100);
    d+=320;
    e+=320;
    memcpy(d, e, 320-100);
  }
}

void sub_bars(void)
{
  #define BARST 15.
  double time=0;
  char sc[64000];
  unsigned char pal[16][3]={
    {255,0,0},    {255-256/4,0,0},    {128,0,0},    {64,0,0}, /* red */
    {0,255,0},    {0,255-256/4,0},    {0,128,0},    {0,64,0}, /* green */
    {0,0,255},    {0,0,255-256/4},    {0,0,128},    {0,0,64}, /* blue */
    {255,255,0},    {255-256/4,255-256/4,0},    {128,128,0},    {64,64,0}  /* yellow */
  };

  memcpy(sc, ecran1, 64000);
  sub_palette((unsigned char *)pal, 16*3, 16);
  set_palette();
  save_palette();

  reset_timer(BARST*1000.);
  while(time<(BARST-2.)/BARST) {
    dobars(time);
    put_screen();
    //memcpy(ecran1, sc, 64000);
    bar_clear_screen(sc);
    time=get_time();
  }
  while(time<1.) {
    dobars(time);
    put_screen();
    //memcpy(ecran1, sc, 64000);
    bar_clear_screen(sc);
    fade_palette(255, 255, 255, 1.-(time-(BARST-2)/BARST)*(BARST/2.), 0, 255);
    set_palette();
    time=get_time();
  }
}

void color_bars(void)
{
  if (show3("graphs/bars1.png", (320-239)/2, 10,
	    "graphs/bars2.png", (320-198)/2+30, 10+35+10+30,
	    "graphs/bars3.png", (320-115)/2+80, 10+35+10+72+10, 13)==-1)
    return;
  pause_me(2.*1000.);
  sub_bars();
}

/******************************************/
/* 2D star field */
void dostar(double time, int pts[4][50][2])
{
  int i, j;
  int dt=time*400.;
  for (i=0; i<4; i++) {
    for (j=0; j<50; j++) {
      int p=pts[i][j][1]*320+(pts[i][j][0]+dt*(i+1))%(320-100)+50;
      ecran1[p]=
      ecran1[p+1]=
      ecran1[p-1]=
      ecran1[p+320]=
      ecran1[p-320]=i+16*3;
    }
  }
}

void sub_star(void)
{
  #define START 15.
  double time=0;
  char sc[64000];
  unsigned char pal[4][3]={
    {0,0,0},
    {155,30,30},
    {200,70,70},
    {255,100,100}
  };
  int pts[4][50][2];
  int i, j;

  memcpy(sc, ecran1, 64000);
  sub_palette((unsigned char *)pal, 16*3, 4);
  set_palette();
  save_palette();

  for (i=0; i<4; i++)
    for (j=0; j<50; j++) {
      pts[i][j][0]=(int)(320.*rand()/(1.+RAND_MAX));
      pts[i][j][1]=50+(int)(100.*rand()/(1.+RAND_MAX));
    }

  reset_timer(START*1000.);
  while(time<(START-2.)/START) {
    dostar(time, pts);
    put_screen();
    memcpy(ecran1, sc, 64000);
    time=get_time();
  }
  while(time<1.) {
    dostar(time, pts);
    put_screen();
    memcpy(ecran1, sc, 64000);
    fade_palette(255, 255, 255, 1.-(time-(START-2)/START)*(START/2.), 0, 255);
    set_palette();
    time=get_time();
  }
}

void star_field(void)
{
  if (show2("graphs/field1.png", (320-303)/2, 10+30,
	    "graphs/field2.png", (320-296)/2, 10+30+35+10+30, 13)==-1)
    return;
  pause_me(2.*1000.);
  sub_star();
}

/******************************************/
/* 3D cube */
void docube(double time)
{
  display_cube(time*30, time*50, time*20, (int)(time*50)%2!=0);
}

void sub_cube(void)
{
  #define CUBET 15.
  double time=0;
  char sc[64000];
  unsigned char pal[6][3]={
    {0,255,0},
    {0,200,0},
    {0,150,0},
    {0,100,0},
    {0,60,0},
    {0,40,0}
  };

  memcpy(sc, ecran1, 64000);
  sub_palette((unsigned char *)pal, 16*3, 6);
  set_palette();
  save_palette();

  init_cube();

  reset_timer(CUBET*1000.);
  while(time<(CUBET-2.)/CUBET) {
    docube(time);
    put_screen();
    memcpy(ecran1, sc, 64000);
    time=get_time();
  }
  while(time<1.) {
    docube(time);
    put_screen();
    memcpy(ecran1, sc, 64000);
    fade_palette(255, 255, 255, 1.-(time-(CUBET-2)/CUBET)*(CUBET/2.), 0, 255);
    set_palette();
    time=get_time();
  }
}

void cube(void)
{
  if (show3("graphs/cube1.png", (320-303)/2, 10,
	    "graphs/cube2.png", (320-300)/2, 10+48,
	    "graphs/cube3.png", (320-225)/2+40, 10+35+10+72+10, 13)==-1)
    return;
  pause_me(2.*1000.);
  sub_cube();
}

/******************************************/
/* thrashy stuff */
point the_points[64+8+32];
facette the_facettes[64+6+32];

void init_palette(void)
{
  int i, j;
  unsigned char pal[256][3];
  pal[0][0]=pal[0][1]=pal[0][2]=0;
  for (i=0; i<8; i++) {
    for (j=0; j<8*2; j++) {
      pal[i*16+j+1][0]=((i<4?i*(32/8):(8-i)*(32/8))+(20+(j>8?(16-j)*(16/8):j*(16/8))))*3/2;
      pal[i*16+j+1][1]=(20+(j>8?(16-j)*(16/8):j*(16/8)));
      pal[i*16+j+1][2]=((i<4?i*(16/8):(8-i)*(16/8))+(20+(j>8?(16-j)*(16/8):j*(16/8))));
    }
  }
  for(i=1; i<129; i++) {
    if (pal[i][0]>63)
      pal[i][0]=63;
    if (pal[i][1]>63)
      pal[i][1]=63;
    if (pal[i][2]>63)
      pal[i][2]=63;
  }
  pal[129][0]=60; pal[129][1]=32; pal[129][2]=42;
  pal[130][0]=63; pal[130][1]=5; pal[130][2]=12;
  pal[131][0]=63; pal[131][1]=63; pal[131][2]=63;
  for (i=0; i<256; i++) {
    pal[i][0]*=4;
    pal[i][1]*=4;
    pal[i][2]*=4;
  }
  sub_palette((unsigned char *)pal, 0, 256);
  set_palette();
  save_palette();
}

void dothrashy(double time)
{
  int a1, a2, a3, i=time*(100*3+25);

  a1=-i*2;
  a2=-i*1;
  a3=-i*3;
  a1&=511;
  a2&=511;
  a3&=511;
  position_points(the_points, a1, a2, a3);
  projete_points(the_points);
  memset(ecran1, 0, 64000);
  draw_facettes(the_points, the_facettes);
}

void dothrashy2(double time)
{
  int a1, a2, a3, i=time*(100*21), nb=100*3+25;

  a1=-nb*2+i*1;
  a2=-nb*1+i*1;
  a3=-nb*3+i*1;
  a1&=511;
  a2&=511;
  a3&=511;
  nb+=i;
  position_points_sexe(the_points, a1, a2, a3, nb);
  projete_points_sexe(the_points);
  memset(ecran1, 0, 64000);
  draw_facettes_sexe(the_points, the_facettes);
}

void dothrashy3(double time)
{
  int a1, a2, a3, i=time*(100*6), nb=100*3+25+21*100;

  a1=-(100*3+25)*2+21*100;
  a2=-(100*3+25)*1+21*100+i*1;
  a3=-(100*3+25)*3+21*100;
  a1&=511;
  a2&=511;
  a3&=511;
  nb+=i;
  position_points_ejacule(the_points, a1, a2, a3, nb);
  projete_points_ejacule(the_points);
  memset(ecran1, 0, 64000);
  draw_facettes_ejacule(the_points, the_facettes, nb);
}

void sub_thrashy(void)
{
  double time=0;
  init_math();
  create_donut(the_points, the_facettes, 8, 8);
  memset(ecran1, 0, 64000);
  init_palette();
  reset_timer(/*3.25*/5*1000.);
  while(time<1.) {
    dothrashy(time);
    put_screen();
    time=get_time();
  }
  time=0;
  reset_timer(/*21.*/5.*1000.);
  while(time<1.) {
    dothrashy2(time);
    put_screen();
    time=get_time();
  }
  time=0;
  reset_timer(/*6.*/5.*1000.);
  while(time<1.) {
    dothrashy3(time);
    put_screen();
    time=get_time();
  }
}

void thrashy(void)
{
  if (show3("graphs/thrashy1.png", (320-228)/2, 10,
	    "graphs/thrashy2.png", (320-186)/2, 10+65,
	    "graphs/thrashy3.png", (320-132)/2, 10+35+10+72+10, 7)==-1)
    return;
  pause_me(7.5*1000.);
  set_pattern(16+4);
  sub_thrashy();
}

/******************************************/
/* vision dedicace */
void put_ass1(int x, int y, char *ass)
{
  int xs=x-100;
  int ys=y-75;
  int dx=200;
  int dy=150;
  int i;
  unsigned char *b=ecran1;

  if (ys<0) {
    ass+=(-ys)*200;
    dy+=ys;
    ys=0;
  }
  if (ys+dy>200) {
    dy=200-ys;
  }
  if (xs<0) {
    ass+=(-xs);
    dx+=xs;
    xs=0;
  }
  if (xs+dx>320) {
    dx=320-xs;
  }

  b+=xs+ys*320;

  for (i=0; i<dy; i++, b+=320, ass+=200)
    memcpy(b, ass, dx);
}

void put_ass2(int x, int y, char *ass)
{
  int xs=x-100;
  int ys=y-75;
  int dx=200;
  int dy=150;
  int i;
  unsigned char *b=ecran1;

  if (ys<0) {
    ass+=(-ys)*200;
    dy+=ys;
    ys=0;
  }
  if (ys+dy>200) {
    dy=200-ys;
  }
  if (xs<0) {
    ass+=(-xs);
    dx+=xs;
    xs=0;
  }
  if (xs+dx>320) {
    dx=320-xs;
  }

  b+=xs+ys*320;

  for (i=0; i<dy; i++, b+=320, ass+=200) {
    unsigned char *from=ass;
    unsigned char *to=b;
    int j, v;
    for (j=0; j<dx; j++, from++, to++) {
      v=*from;
      if (v>(*to&63))
	*to=v;
    }
  }
}

void dovision(double time, char *ass)
{
  int y1=cos(time*8*M_PI)*30.+100.;
  int y2=cos(time*8*M_PI+M_PI/2.)*30.+100.;

  put_ass1(160-30, y1, ass);
  put_ass2(160+30, y2, ass);
}

void dovision2(double time, char *ass)
{
  int y1=cos(time*4*M_PI)*30.+100.;
  int y2=cos(time*4*M_PI+M_PI/2.)*30.+100.;
  int y=(y1+y2)/2;
  int x=160-30+time*100.;

  if (x>160-30+50) {
    x=160-30+50;
    y=111+(time-0.5)*300;
    put_ass1(160+30, y2, ass);
    put_ass2(160-30, y1, ass);
    fill_triangle_8bpp(x, y, x, y+20, x+20, y+10, 64);
    return;
  }
  put_ass1(160+30, y2, ass);
  fill_triangle_8bpp_2(x, y, x, y+20, x+20, y+10, 64);
  put_ass2(160-30, y1, ass);
}

void vision_palette()
{
  int i;
  unsigned char pal[65][3];
  memset(ecran1, 0, 64000);

  uni_palette(181, 102, 18);
  for (i=0; i<64; i++) {
    pal[i][0]=i*4;
    pal[i][1]=0;
    pal[i][2]=0;
  }
  sub_palette((unsigned char *)pal, 0, 64);
  set_palette();
  save_palette();
  put_screen();
}

void create_ass(char *ass)
{
  int i, j, v;
  double dmax=100.;//sqrt(100.*100.+75.*75.);
  for (i=0; i<200; i++)
    for (j=0; j<150; j++) {
      v=sqrt(((double)i-100.)*((double)i-100.)+(((double)j-75.)*200./150.)*(((double)j-75.)*200./150.))*64./dmax;
      v=63-v;
      if (v<0)
	v=0;
      ass[j*200+i]=v;
    }
}

void sub_vision(void)
{
  int i;
  double time;
  char ass[200*150];

  vision_palette();
  create_ass(ass);

  for (i=0; i<6; i++) {
    time=0;
    reset_timer(2*1000.);
    while(time<1.) {
      memset(ecran1, 0, 64000);
      dovision(time, ass);
      put_screen();
      time=get_time();
    }
    time=0;
    reset_timer(2*1000.);
    while(time<1.) {
      memset(ecran1, 0, 64000);
      dovision2(time, ass);
      put_screen();
      time=get_time();
    }
  }
}

void vision(void)
{
  if (show3("graphs/vision1.png", (320-258)/2, 10,
	    "graphs/vision2.png", (320-261)/2, 10+65,
	    "graphs/vision3.png", (320-180)/2, 10+35+10+72+10, 7)==-1)
    return;
  pause_me(5.*1000.);
  set_pattern(13);
  sub_vision();
}

/******************************************/
/* black screen of death */
void death(void)
{
  if (show4("graphs/death1.png", 10, 2,
	    "graphs/death2.png", 2, 10+40,
	    "graphs/death3.png", (320-231)/2, 10+65+20+10,
	    "graphs/death4.png", (320-252)/2, 10+35+10+72+10+10, 7)==-1)
    return;
  pause_me(5.*1000.);
  uni_palette(0, 0, 0);
  set_palette();
  pause_me(10.*1000.);
}

/******************************************/
/* end part */
void end1(void)
{
  double time=0.;
  if (show3("graphs/loser1.png", 10, 10,
	    "graphs/loser2.png", (320-261)/2+20, 10+65+5,
	    "graphs/loser3.png", (320-180)/2, 10+35+10+72+10, 8)==-1)
    return;
  pause_me(7.*1000.);
  reset_timer(3.*1000.);
  while(time<1.) {
    fade_palette(255, 255, 255, 1.-time, 0, 255);
    set_palette();
    time=get_time();
  }
}

void end2(void)
{
  double time=0.;
  if (show2("graphs/sed1.png", 2, 10+10+20,
	    "graphs/sed2.png", (320-292)/2, 10+10+10+72, 13)==-1)
    return;
  pause_me(7.*1000.);
  reset_timer(3.*1000.);
  while(time<1.) {
    fade_palette(255, 255, 255, 1.-time, 0, 255);
    set_palette();
    time=get_time();
  }
}

void end3(void)
{
  double time=0.;
  if (show4("graphs/penguin.png", 320-102-10, 10+50,
	    "graphs/boot1.png", 5, 10+60,
	    "graphs/boot2.png", 5, 10+25+5+60+10,
	    "graphs/boot3.png", 5+80, 10+35+10+70+10+15, 7)==-1)
    return;
  pause_me(7.*1000.);
  reset_timer(3.*1000.);
  while(time<1.) {
    fade_palette(255, 255, 255, 1.-time, 0, 255);
    set_palette();
    time=get_time();
  }
}

void end(void)
{
  end1();
  end2();
  end3();
}
