/* this file is public domain */

#include "sound.h"

#include <signal.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>

#include "mixer.h"
#include "xm.h"
#include "card.h"

#include <string.h>
#include <stdio.h>

int sok=0;
XM the_xm;

static volatile int critik_section=0;
void enter_sound_critik(void)
{
  while(critik_section);
  critik_section=1;
}

void leave_sound_critik(void)
{
  critik_section=0;
}

/* 1 if yes, 0 if no */
int can_mix(int fd)
{
  fd_set r, w, e;
  struct timeval t;

  FD_ZERO(&r);
  FD_ZERO(&w);
  FD_ZERO(&e);

  t.tv_sec=0;
  t.tv_usec=0;

  FD_SET(fd, &w);

  select(FD_SETSIZE, &r, &w, &e, &t);
  return FD_ISSET(fd, &w);
}

void rubick_mix_timer(int n)
{
  if (!sok)
    return;

  enter_sound_critik();
  while (can_mix(get_sound_fd()))
   mix(&the_xm);
  leave_sound_critik();
}

/* return 1 if ok, 0 otherwise */
int init_timer(void)
{
  struct sigaction act;
  struct itimerval tim;

  act.sa_handler=rubick_mix_timer;
  sigemptyset(&act.sa_mask);
  act.sa_flags=0;
  act.sa_restorer=0;
  if(sigaction(SIGALRM, &act, (struct sigaction *)0)==-1) {
    fprintf(stderr, "error with sigaction, sorry...\n");
    return 0;
  }

  /* a timer to 50 Hz */
  tim.it_interval.tv_sec=0;
  tim.it_interval.tv_usec=20*1000;
  tim.it_value.tv_sec=0;
  tim.it_value.tv_usec=20*1000;
  if(setitimer(ITIMER_REAL, &tim, (struct itimerval *)0)==-1) {
    fprintf(stderr, "error with setitimer, sorry...\n");
    return 0;
  }

  sok=1;
  return 1;
}

/* -1 if bug, 0 if good */
int rubick_init_sound(char *file)
{
  fprintf(stderr, "Initing the sound...\n");
  if (!init_sound(44100, 16, 1)) {
    fprintf(stderr, "Error with initing the sound !\n");
    return 0;
  }
  fprintf(stderr, "Sound inited ok.\n");

  memset(&the_xm, 0, sizeof(XM));

  fprintf(stderr, "Loading the module \"%s\"...\n", file);
  if (!load_xm(file, &the_xm)) {
    fprintf(stderr, "Error with the loading of \"%s\"\n", file);
    close_sound();
    return 0;
  }
  fprintf(stderr, "File \"%s\" loaded ok.\n", file);

  init_xm_data(&the_xm);
  the_xm.mix=MODULE_PLAY;

  return init_timer();
}

void rubick_close_sound(void)
{
  if (!sok)
    return;

  enter_sound_critik();
  close_sound();
  leave_sound_critik();
}

/* set the pattern to the value given */
void set_pattern(int pat)
{
  int i;
  if (!sok)
    return;

  enter_sound_critik();
  close_all_channels(&the_xm);
  for (i=0; i<the_xm.nb_channels; i++)
    the_xm.channels[i].loop_point_is_set=0;
  the_xm.current_pattern=pat;
  the_xm.current_line=-1;
  the_xm.speed=32768;
  the_xm.to_handle=new_line(&the_xm);
  leave_sound_critik();
}
